/*
   B5TOUNI - A BIG5 to Unicode SGML character reference converter
   Copyright (C) 2001 Chih-Hao Tsai

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
 */

/*
   This program requires the "BIG5 to Unicode Table" compiled by
   Glenn Adams <glenn@metis.com> and John H. Jenkins 
   <John_Jenkins@taligent.com>. It can be downloaded at the
   following location:

   ftp://ftp.unicode.org/Public/MAPPINGS/EASTASIA/OTHER/BIG5.TXT
*/

/*
   Module: b5touni.c
   Version: 0.10
   Date: April 13, 2001
   Author: Chih-Hao Tsai
   Email: hao520@yahoo.com
   B5TOUNI website: http://www.geocities.com/hao510/b5touni/
   Author's homepage: http://www.geocities.com/hao520/
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define DELIMITER " \n\r\t"
#define TABLE "big5.txt"
#define TABLE_SIZE 14000


typedef struct buxref BUxref;
struct buxref
{
  char b5[3];
  int unicode;
};

int qsort_helper (const void *, const void *);
int is_big5_leading_char (const unsigned char);
int is_big5_trailing_char (const unsigned char);

int
main (int argc, char **argv)
{
  FILE *B5TOUNI;
  BUxref *BU, *BP;
  char string[256];
  char *filename;
  char *u, *b5;
  char *bp;
  char temp0, temp1, b5char[3];
  int unicode, big5;
  int char_count = 0;

  filename = argc > 1 ? argv[1] : TABLE;


  BU = (BUxref *) malloc (sizeof (BUxref) * TABLE_SIZE);
  memset (BU, 0, sizeof (BUxref) * TABLE_SIZE);
  memset (b5char, 0, sizeof (b5char));

  if ((B5TOUNI = fopen (filename, "rt")))
    {
      while (fgets (string, 255, B5TOUNI))
	{
	  if (b5 = strtok (string, DELIMITER))
	    if (strlen (b5) == 6)
	      {
		if (u = strtok (0, DELIMITER))
		  if (strlen (u) == 6)
		    {
		      unicode = (int) strtol (u, &bp, 0);
		      big5 = (int) strtol (b5, &bp, 0);
		      BU[char_count].b5[0] = (int) big5 / 256;
		      BU[char_count].b5[1] = (int) big5 % 256;
		      BU[char_count].unicode = unicode;
		      char_count++;
		    }
	      }
	}
      fclose (B5TOUNI);
    }
  else
    {
      fprintf (stderr, "Unable to load mapping table.\n");
      exit (1);
    }

  qsort (BU, char_count, sizeof (BUxref), qsort_helper);

  while ((temp0 = fgetc (stdin)) != EOF)
    {
      if (!is_big5_leading_char (temp0))
	fputc (temp0, stdout);
      else
	{
	  if ((temp1 = fgetc (stdin)) != EOF)
	    if (is_big5_trailing_char (temp0))
	      {
		b5char[0] = temp0;
		b5char[1] = temp1;
		BP =
		  bsearch (b5char, BU, char_count, sizeof (BUxref),
			   qsort_helper);
		if (BP)
		  fprintf (stdout, "&#%d;", BP->unicode);
		else
		  {
		    fputc (temp0, stdout);
		    fputc (temp1, stdout);
		  }
	      }
	    else
	      {
		fputc (temp0, stdout);
		fputc (temp1, stdout);
	      }

	}
    }

  free (BU);
  return (0);
}


int
qsort_helper (const void *e1, const void *e2)
{
  return strcmp (((BUxref *) e1)->b5, ((BUxref *) e2)->b5);
}

int
is_big5_leading_char (const unsigned char leading)
{
  if (!leading)
    return (0);
  else if (leading >= 0xa1 && leading <= 0xf9)
    return (1);
  else
    return (0);
}

int
is_big5_trailing_char (const unsigned char trailing)
{
  if (!trailing)
    return (0);
  else if ((trailing >= 0x40 && trailing <= 0x7e)
	   || (trailing >= 0xa1 && trailing <= 0xfe))
    return (1);
  else
    return (0);
}
